/*
//  File  : matrix.c
//  (c)HOOPOE ProGroup 2011
*/

/*
// C= Programming Sample
*/

/// Copyright string
char* copyright = "(c)HOOPOE ProGroup 2011";

#include <stdio.h>
#include <C=.h>

#include <intrin.h>
#pragma intrinsic(__rdtsc)


/// input/output matrices

#define L 1250
#define M 1310
#define N 1470

float a[L][M];
float b[N][L];
/// c = a * b;
float c[N][M];
float d[N][M];
float e[N][M];

void main()
{
    /// TODO: Illustrate data-decomposition parallelism and dynamic load balancing
    ///       in the example of matrix multiplication

    __int64 seqtime;
    __int64 partime;
    __int64 dyntime;

    int i, j, k;

    /// initialize input matrices
    for(j = 0; j < L; j++)
    {
        for(i = 0; i < M; i++)
        {
            a[j][i] = (float)(i + j * M);
        }
    }
    for(j = 0; j < N; j++)
    {
        for(i = 0; i < L; i++)
        {
            b[j][i] = (float)(i + j * L);
        }
    }

    seqtime = __rdtsc();

    /// multiply the matrices sequentially
    for(j = 0; j < N; j++)
    {
        for(i = 0; i < M; i++)
        {
            c[j][i] = 0.f;

            for(k = 0; k < L; k++)
            {
                c[j][i] += a[k][i] * b[j][k];
            }
        }
    }
    seqtime = __rdtsc() - seqtime;

    partime = __rdtsc();

    /// multiply the matrices in parallel, static decomposition 
    parallel()
    {
        int end = CPX_RANGE_END(N);

        for(j = CPX_RANGE_START(N); j < end; j++)
        {
            for(i = 0; i < M; i++)
            {
                d[j][i] = 0.f;

                for(k = 0; k < L; k++)
                {
                    d[j][i] += a[k][i] * b[j][k];
                }
            }
        }
        barrier();
    }

    partime = __rdtsc() - partime;

    dyntime = __rdtsc();

    /// multiply the matrices in parallel, dynamic balancing
    parallel(N)
    {
        j = pix();

        for(i = 0; i < M; i++)
        {
            e[j][i] = 0.f;

            for(k = 0; k < L; k++)
            {
                e[j][i] += a[k][i] * b[j][k];
            }
        }
        barrier();
    }

    dyntime = __rdtsc() - dyntime;

    /// print timings
    printf("Serial:   %llu\n", seqtime);
    printf("Parallel: %llu\n", partime);
    printf("Dynamic:  %llu\n", dyntime);

    for(j = 0; j < N; j++)
    {
        for(i = 0; i < M; i++)
        {
            if(c[j][i] != d[j][i] || c[j][i] != e[j][i])
            {
                printf("Error!\n");
                goto exit;
            }
        }
    }
    printf("OK\n");

exit:;

    /*
    /// print the results
    for(j = 0; j < N; j++)
    {
        for(i = 0; i < M; i++)
        {
            printf("%f\t", c[j][i]);
        }
        printf("\n");
    }
    */
}
